(*The file illustrates some inductive and coinductive definitions; and 
also shows the Coq syntax for recursive and corecursive functions.
The file was used to provide a short introduction to Coq capacities for
my talk in Marseille, 6th of May 2008*)




(*Coinductive Types and Corecursive Functions*)

Section Coinduction.

Variable A : Set.

CoInductive Stream : Set :=
    Cons : A -> Stream -> Stream.

Set Implicit Arguments.

CoInductive LList (A:Set) : Set :=
LNil : LList A
|LCons : A -> LList A -> LList A.

Implicit Arguments LNil [A].

CoFixpoint from (n: nat) : LList nat := LCons n (from (S n)).

Definition Nats : LList nat := from 0.

Fixpoint LNth (A : Set) (n: nat) (l: LList A) {struct n}:
option A :=
match l with 
| LNil => None
| LCons a l' => match n with O => Some a 
                             | S p => LNth p l' end
end.



(*------------------------------------------*)

Definition LList_decompose (A: Set) (l: LList A) : LList A :=
match l with
| LNil => LNil
| LCons a l' => LCons a l'
end.

Theorem LList_decomposition_lemma :
forall (A : Set) (l: LList A), l = LList_decompose l.
Proof.
intros; case l. 
trivial.
intros.
unfold LList_decompose.
trivial.
Qed.

Ltac LList_unfold term :=
apply trans_equal with (1 := LList_decomposition_lemma term).

(*------------------------------------------*)



CoFixpoint repeat (A : Set) (a : A) : LList A :=
LCons a (repeat a).

Lemma repeat_unfold : forall (A: Set)(a: A), 
repeat a = LCons a (repeat a).
Proof.
intros.
LList_unfold (repeat a).
simpl; auto.
Qed.


(*Inductive Predicates over CoInductive Types*)

Inductive Finite (A: Set): LList A -> Prop :=
|Finite_LNil: Finite LNil
|Finite_LCons: forall (a : A) (l: LList A), 
Finite l -> Finite (LCons a l). 

(*CoInductive Predicates*)


CoInductive Infinite (A : Set): LList A -> Prop :=
Infinite_LCons: forall (a : A) (l: LList A), 
Infinite l -> Infinite (LCons a l).


(* Building Infinite Proofs:
Cofix: Coinductive form of Induction principle. 
Attention to Guardedness *)

Lemma repeat_infinite : forall (A: Set) (a: A), 
Infinite (repeat a).
Proof.
cofix H.
(*
assumption.
Qed.*)
intros.
rewrite (repeat_unfold).
apply Infinite_LCons.
apply H.
Qed.


(*Useful tactic "inversion"*)

Theorem LNil_not_Infinite : forall (A :Set), 
~ Infinite (LNil (A := A)).
Proof.
intros A0 H. 
inversion H.
Qed.

Theorem Infinite_of_LCons :
forall (A : Set) (a : A) (u: LList A),
Infinite (LCons a u) -> Infinite u.
Proof.
intros A0 a u H0. 
inversion H0.
exact H1.
Qed.

(*Bisimilarity*)

CoInductive bisimilar (A :Set) : LList A -> LList A -> Prop :=
|bisim0: bisimilar LNil LNil
|bisim1: forall (a : A) (l l' : LList A), 
bisimilar l l' -> bisimilar (LCons a l)(LCons a l').


Theorem Bisim_refl: forall (A :Set)(l : LList A), 
bisimilar l l.
Proof.
cofix H.
intros.
case l.
apply bisim0.
intros.
apply bisim1.
apply H.
Qed.

Theorem Bisim_trans: forall (A : Set) (l u v: LList A),
 bisimilar  l u -> bisimilar u v -> bisimilar l v.
Proof.
cofix H.
intros A0 l u v.
case l; case u; case v. 
intros; constructor.
intros a l0 H0 H1; exact H1.
intros; constructor.
intros a l0 a0 l1 H0 H1.
inversion H0.
intros a l0 H0 H1; exact H0.
intros a l0 a0 l1 H0 H1.
inversion H0.
intros a l0 a0 l1 H0 H1.
inversion H1.
intros a l0 a0 l1 a1 l2 H0 H1.
inversion H0.
inversion H1.
constructor.
generalize H8; generalize H3.
apply H.
Qed.


Theorem Bisim_sym: forall (A:Set)(l u :LList A), 
bisimilar l u -> bisimilar u l.
Proof.
cofix H.
intros A0 l u.
case l; case u.
intro; assumption.
intros a l0 H0; inversion H0.
intros a l0 H0; inversion H0.
intros a l0 a0 l1 H0; inversion H0.
constructor.
generalize H2.
apply H.
Qed.

Print Bisim_sym.

Section Coinductive_Proof_Principle.
(* Jacobs, Rutten (A Tutorial on Coalgebras and Coinduction):
CPP: For all infinite lists u, v,  bisimilar u v -> u = v.
Proof is based on finality of the coalgebra of infinite 
lists of elements of A.*)

(*Constructive version:*)

Lemma bisimilar_LNth :
forall (A : Set) (n: nat) (u v: LList A),
bisimilar u v -> LNth n u = LNth n v.
Proof.
induction n.
(*Basic step*)
intros u v; case u; case v; auto.
intros a l H; inversion H.
intros a l H; inversion H.
intros a l a0 l0 H; inversion H; auto.
(*Inductive step*)
intros u v; case u; case v. 
intro; auto.
intros; inversion H.
intros; inversion H.
intros; inversion H.
simpl.
apply IHn; exact H1.
Qed.

(*Another constructive version*)

Theorem bisimilar_of_Finite_is_eq:
forall (A : Set) (l: LList A), 
Finite l -> forall l':LList A, bisimilar l l' -> l = l'.
Proof.
intros A0 l H; induction H; intro l'; case l'.
(*Basis step: Finite LNil*)
intro; trivial.
intros a l H; inversion H.
intro H0; inversion H0.
(*Inductive step: Finite l -> Finite LCons a l.*)
intros.
inversion H0.
assert (forall (A :Set) (a: A) (l l': LList A), 
l = l' -> LCons a l = LCons a l').
intros A1 a2 l2 l'1 H6; induction H6; trivial.
apply H6.
generalize H2.
apply IHFinite.
Qed.


(*induction -> elim; case -> destruct*)
 
Lemma LNth_bisimilar :
forall (A : Set) (u v : LList A),
(forall (n : nat), LNth n u = LNth n v) -> bisimilar u v.
Proof.
intro A0; cofix H.
 simple destruct u; simple destruct v.
 intro; constructor.
 intros a l H1.
 generalize (H1 0); discriminate 1.
(*intro H2. discriminate H2.*)
 intro H1.
 generalize (H1 0); discriminate 1.
 intros.
 generalize (H0 0); simpl in |- *.
 injection 1.
 simple induction 1.
 constructor.
 Guarded.
 apply H.
 intro n.
 generalize (H0 (S n)).
 simpl in |- *. 
 auto.
Qed.




Theorem bisimilar_of_Finite_is_Finite:
forall (A : Set) (l : LList A),
Finite l -> forall (l' : LList A), bisimilar l l' -> Finite l'.
Proof.
intros A0 l H; induction H.
(*Basis step*)
intro l'; case l'.
intro; constructor.
intros.
inversion H.
(*Inductive step*)
intro l'; case l'.
intro H1; inversion H1.
intros.
inversion H0.
constructor.
auto.
(*auto uses inductive assumption*)
Qed.

Theorem bisimilar_of_Infinite_is_Infinite : 
forall (A : Set) (l : LList A),
Infinite l -> 
forall l': LList A, bisimilar l l' -> Infinite l'.
Proof.
cofix H.
intros A0 l; case l.
intros H0 l'; case l'.
intro; exact H0.
intros.
inversion H1.
intros a l0 H0 l'; case l'.
intro H1; inversion H1.
intros.
inversion H1.
constructor.
generalize H3.
apply H.
Guarded.
apply Infinite_of_LCons with (a := a).
exact H0.
Qed.

End Coinductive_Proof_Principle.

Hint Resolve Finite_LNil Finite_LCons : llists.

Hint Resolve Infinite_LCons: llists.

Hint Resolve bisim0 bisim1: llists.

Section LTL.
(*A. Pnueli. The temporal semantics of concurrent programs. 
TCS (31): 45-60, 1981.
Also: B. Jacobs "Exercises in Coalgebraic Specification"*)

Variables (a b c : A).

Definition satisfies (l: LList A) (P : LList A -> Prop) : 
Prop := P l.

Hint Unfold satisfies: llists.



Inductive Eventually (P : LList A -> Prop) : LList A -> Prop :=
| Eventually_here: 
forall (a : A)(l: LList A), 
P (LCons a l) -> Eventually P (LCons a l)
| Eventually_further : 
forall (a : A) (l: LList A), 
Eventually P l -> Eventually P (LCons a l).

Hint Resolve Eventually_here Eventually_further.



CoInductive Always (P : LList A -> Prop) : LList A -> Prop :=  
Always_Lcons: 
forall (a : A) (l: LList A),
P (LCons a l) -> Always P l -> Always P (LCons a l).

Theorem Always_Infinite:
forall (l : LList A) (P: LList A -> Prop), 
satisfies l (Always P) -> Infinite l.
Proof.
unfold satisfies.
cofix H.
intros l P H0; case H0.
intros.
constructor.
generalize H2.
apply H.
Qed.



Definition F_infinite (P : LList A  -> Prop) : 
LList A -> Prop :=
Always (Eventually P).


Definition F_Infinite_2_Eventually: forall P l, 
F_infinite P l -> Eventually P l.
intros P l. 
destruct l.
intro.
inversion H.
intro.
inversion H; assumption.
Defined.

End LTL.
End Coinduction.























 









